-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}

M.wheels = {}
M.damage = 0
M.lowpressure = false

local totalBreakEnergy = 0
local breakNode = -1
local maxEnergy = 0
local collTriState = {}

local materialBrokeQueue = {}
local prevElapsedTime = 0

local unstableWheels = {}

--Dummiesman's regeneration mod
--AMOUNT TO REGEN EACH TIME
local regenrate = 0.0025
--REGENERATING?
local regenerating = false
--USER FRIENDLY.. :D (kinda)
local laststatus = "Not regenerating"

-- local helper function
local function luaBreakBeam(id)
    -- notify the rest
    drivetrain.beamBroke(id)
    electrics.beamBroke(id)
end

local function deflateTire(wheelid, energy)
    local wheel = M.wheels[wheelid]
    local i, beam
    M.lowpressure = true
    
    --if wheel.reinforcementBeams ~= nil then
    --    for i, beam in pairs (wheel.reinforcementBeams) do
    --        obj:setBeamSpringDamp(beam.cid, beam.beamSpring * 0.5, -1, beam.beamSpring * 0.02, -1)
    --    end
    --end

    if wheel.pressureGroup ~= nil then
        if v.data.pressureGroups[wheel.pressureGroup] ~= nil and unstableWheels[wheel.pressureGroup] == nil then
            if energy ~= nil and energy > 1000 then 
                obj:deflatePressureGroup(v.data.pressureGroups[wheel.pressureGroup])
                unstableWheels[wheel.pressureGroup] = 1
            else
                obj:setGroupPressure(v.data.pressureGroups[wheel.pressureGroup], (0.2 * 6894.757 + 101325))
            end
        end
    end
    
    if wheel.treadBeams ~= nil then
        for i, beam in pairs (wheel.treadBeams) do
            obj:setBeamSpringDamp(beam.cid, beam.beamSpring * 0.0, 0, -1, -1)
        end
    end
    
    if wheel.sideBeams ~= nil then
        for i, beam in pairs (wheel.sideBeams) do
            obj:setBeamSpringDamp(beam.cid, beam.beamSpring * 0.0, 0, -1, -1)
        end
    end

    if wheel.reinfBeams ~= nil then
        for i, beam in pairs (wheel.reinfBeams) do
            obj:setBeamSpringDamp(beam.cid, 0, 0, 0, 0)
        end
    end

    if wheel.pressuredBeams ~= nil then
        for i, beam in pairs (wheel.pressuredBeams) do
            obj:setBeamPressureRel(beam.cid, 0, math.huge, -1, -1)
        end
    end
end

local function getBeamPartOrigin(b)
	local x = v.data.nodes[b.id1].partOrigin
	if x == nil then x = "" end
	return x
end
local function getBeamPartOrigin2(b)
	local x = v.data.nodes[b.id2].partOrigin
	if x == nil then x = "" end
	return x
end
local function validRegenPart(po)
	local isValid = true
	if string.find(po,"coilover") or string.find(po,"suspension") or  string.find(po,"steering") or  string.find(po,"swaybar") or  string.find(po,"brake") or string.find(po,"spring") or string.find(po,"wheel") then isValid = false end
	return isValid
end

local function scaleThresh()
	return regenrate * (bullettime.simulationSpeed/100)
end

local function doRegen()
	if regenerating then
		local thrsh = scaleThresh()
		local numFull = 0
		local numTotal = 0
		for k,beam in pairs(v.data.beams) do
			local po = getBeamPartOrigin(beam)
			local isValid = validRegenPart(po)
			po = getBeamPartOrigin2(beam)
			local isValid2 = validRegenPart(po)
			
		    if (isValid and isValid2) and not obj:beamIsBroken(beam.cid) then
				numTotal = numTotal + 1
				local cr = obj:getBeamLengthRefRatio(k)
				
				local lleft = math.abs(cr - 1)
				if  lleft < 0.005 then
					numFull = numFull + 1
				else
					if cr < 1 then
						obj:setBeamLengthRefRatio(k,cr + thrsh)
					elseif cr > 1 then
						obj:setBeamLengthRefRatio(k,cr - thrsh)
					end
				end
			end
		end
		if numFull == numTotal then 
          regenerating = false 
          laststatus = "Regeneration Complete!" 
          material.reset()
        else
          laststatus = "regeneration progress " .. numFull .. "/" .. numTotal 
        end
	end
end
local function updateGFX()
    if regenerating then doRegen() end
    -- crash sounds
    if totalBreakEnergy ~= 0 then
        if breakNode ~= -1 then
            sounds.playSoundOnceAtNode("CrashTestSound", breakNode, math.log10(totalBreakEnergy))
            prevElapsedTime = 0
            M.damage = M.damage + totalBreakEnergy
        end
        totalBreakEnergy = 0
        breakNode = -1
        maxEnergy = 0
    end
    
    -- work off the material switching, needs to happen in synchronous mode
    if obj.ibody ~= nil then
        for k,beam in pairs(materialBrokeQueue) do
            material.switchBrokenMaterial(beam.deformSwitches)
            M.damage = M.damage + 100
        end
    end
    materialBrokeQueue = {}
end

local function beamBroken(id, energy)
    local beam

    if energy > maxEnergy then
        breakNode = v.data.beams[id].id1
        maxEnergy = energy
    end
    totalBreakEnergy = totalBreakEnergy + energy

    luaBreakBeam(id)
    if v.data.beams[id] ~= nil then
        local beam = v.data.beams[id]
        
        if bdebug.enabled == true then
            local m ="beam "..id.." just broke: " .. v.data.nodes[beam.id1].name .. " ["..beam.id1.."]" .."  ->  " .. v.data.nodes[beam.id2].name .. " ["..beam.id2.."]"
            log('D', "beamstate.beamBroken", m)
            gui.message(m)
        end
        
        -- Break coll tris
        if beam.collTris then
            for k,ctid in pairs(beam.collTris) do
                if collTriState[ctid] then
                    --collTriState[ctid] = collTriState[ctid] - 1
					collTriState[ctid] = 0
                    if collTriState[ctid] <= 1 then
                        obj:breakCollisionTriangle(ctid)
                    end
                end
            end
        end
        
        -- Break the meshes
        if beam.disableMeshBreaking == nil or beam.disableMeshBreaking == false then
            obj:breakMeshes(id)
        end

        -- Break rails
        obj:breakRails(id)
        
        -- Check for punctured tire
        if beam.wheelID ~= nil then
            M.wheels[beam.wheelID].punctured = true
            deflateTire(beam.wheelID, energy)
        end
        
        -- breakgroup handling
        if beam.breakGroup and beam.breakGroup ~= "" then
            -- find all beams with that breakgroup
            if beam.breakGroupType == 0 or beam.breakGroupType == nil then
                -- break all beams in that group
                for k, b in pairs(v.data.beams) do
                    if b.breakGroup == beam.breakGroup then
                        --log('D', "beamstate.beamBroken","  breakgroups: also breaking beam "..k.. " as in breakgroup ".. b.breakGroup)
                        obj:breakBeam(k)
                        luaBreakBeam(k)
                    end
                end

            elseif beam.breakGroupType == 1 then
                -- this type does not break the group but will be broken by the group
            end
        end
        
        if beam.deformSwitches then
            table.insert(materialBrokeQueue, beam)
        end
    else
        --print ("beam "..id.." just broke")
    end
    
    -- funky: disable engine once a beam breaks ;)
    --BeamEngine:setEnabled(false)
end

local function init()
    totalBreakEnergy = 0
    breakNode = -1
    maxEnergy = 0
    M.damage = 0
    materialBrokeQueue = {}
    unstableWheels = {}

    if v.data.wheels then
        M.wheels = deepcopy(v.data.wheels)
        for i, wheel in pairs(M.wheels) do
            wheel.punctured = false
            wheel.pressureCoef = 1
        end
    end    
    
    -- Reset colltris
    if v.data.triangles then
        collTriState = {}
        for k, v in pairs(v.data.triangles) do
            if v.cid and v.beamCount then
                collTriState[v.cid] = v.beamCount
            end
        end
    end
end

local function beamDeformed(id, ratio)
    --log('D', "beamstate.beamDeformed","beam "..id.." deformed with ratio "..ratio)
    if v.data.beams[id] then
        local beam = v.data.beams[id]
        if beam.deformSwitches then
            table.insert(materialBrokeQueue, beam)
        end
    end
end

local function reset()
    init()
    M.lowpressure = false
end

local function breakAllBreakgroups()
    for k, b in pairs(v.data.beams) do
        if b.breakGroup ~= nil and b.breakGroup ~= "" then
            obj:breakBeam(k)
        end
    end
end

local function breakHinges()
    for k, b in pairs(v.data.beams) do
        if b.breakGroup ~= nil and b.breakGroup ~= "" then
            if string.find(b.breakGroup, "hinge") ~= nil or string.find(b.breakGroup, "latch") ~= nil then
                --log('D', "beamstate.breakHinges","  breaking hinge beam "..k.. " as in breakgroup ".. b.breakGroup)
                obj:breakBeam(k)
            end
        end
    end
end

local function deflateTires()
    for i, wheel in pairs(M.wheels) do
        wheel.punctured = true
        deflateTire(i, 0)
        M.damage = M.damage + 1000
    end
end

local function breakBreakGroup(group)
    for k,v in pairs(v.data.beams) do
        if v.breakGroup == group then 
            obj:breakBeam(v.cid)
        end
    end
end

local function addDamage(damage)
    M.damage = M.damage + damage
end

local function regenerate()
    regenerating = true
    local i,beam
    for i, beam in pairs (v.data.beams) do
        obj:setBeamSpringDamp(beam.cid, beam.beamSpring or -1, beam.beamDamp or -1, beam.springExpansion or -1, beam.dampExpansion or -1)
    end
end

local function stopRegenerate()
	regenerating = false
	laststatus = "Regen manual stop"
end

local function getRegenStatusUI()
	local returnTable = {}
	returnTable.text = laststatus
	return returnTable
end

-- public interface
M.beamBroken  = beamBroken
M.reset       = reset
M.init          = init
M.deflateTire = deflateTire
M.updateGFX   = updateGFX
M.beamDeformed = beamDeformed
M.breakAllBreakgroups = breakAllBreakgroups
M.breakHinges = breakHinges
M.deflateTires = deflateTires
M.breakBreakGroup = breakBreakGroup
M.addDamage = addDamage
M.regenerate = regenerate
M.stopRegenerate = stopRegenerate
M.regenerationRate = regenrate
M.getRegenState = getRegenStatusUI

return M